import pandas as pd


def convert_df_to_block(dataframe):
    return PandasBlock(dataframe)


def build_block_from_rows(rows):
    df = pd.concat(rows)
    return PandasBlock(df, relationship=None, subject=None, origin=None)


def build_new_learnblock(values, columns, index, origin):
    data_frame = pd.DataFrame(data=values, columns=columns, index=index)
    block = PandasBlock(data_frame, origin=origin)
    return block


class PandasBlock:
    LAST_THREE_COLUMNS = 3
    TIME = "T"
    SUBJECT = "Sigma"
    PURPOSE = "Z"

    def __init__(self, dataframe, relationship=None, subject=None, origin=None):
        self.df = dataframe
        self._origin = origin
        self.relationship = relationship
        self.subject = subject
        # TODO (dmt): Check if you really need n_cluster.
        self.n_cluster = None

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        assert isinstance(value, tuple)
        self._origin = value

    def __str__(self):
        return ("origin: <{}> subject: <{}> "
                "n_cluster: <{}> relationship: <{}>".format(self.origin,
                                                            self.subject,
                                                            self.n_cluster,
                                                            self.relationship))

    def __len__(self):
        return self.n_rows()

    def __getitem__(self, item):
        end = self.n_columns(effective=True)
        return self.df.iloc[item][:end]

    @classmethod
    def read_csv(cls, path):
        pb = PandasBlock(dataframe=pd.read_csv(path, index_col=False),
                         origin=("source",))
        columns = pb.columns()
        if (cls.TIME in columns) and (cls.SUBJECT in columns) and \
           (cls.PURPOSE in columns):
            return pb
        raise ValueError(
            ("Tried to read the csv file {}. Could not found "
             "columns {} or {} or {}. Make sure the csv file "
             "contains them.".format(path, cls.TIME, cls.SUBJECT, cls.PURPOSE))
        )

    @classmethod
    def read_table(cls, path):
        return PandasBlock(pd.read_table(path, index_col=False),
                           origin=("source", ))

    @classmethod
    def read_excel(cls, path):
        return PandasBlock(pd.read_excel(path, index_col=False),
                           origin=("source", ))

    def row_as_df_via_index(self, index):
        return self.df.loc[[index]]

    def n_rows(self):
        return self.df.shape[0]

    def n_columns(self, effective=False):
        if effective:
            return max(0, self.df.shape[1] - self.LAST_THREE_COLUMNS)
        else:
            return self.df.shape[1]

    def columns(self, effective=False):
        if not effective:
            return list(self.df.columns)
        else:
            features = list(self.df.columns)
            features.remove(self.TIME)
            features.remove(self.PURPOSE)
            features.remove(self.SUBJECT)
            return features

    def min_timestamp(self):
        return min(self.df["T"])

    def max_timestamp(self):
        return max(self.df["T"])

    def indices(self):
        return list(self.df.index)

    def reindex(self, indices):
        assert len(indices) == self.df.shape[0]
        self.df.index = indices
        # self.df.reindex(indices)

    def has_nan(self):
        return bool(self.df.isna().any().any())

    def drop_rows(self, rows):
        self.df.drop(rows, inplace=True)

    def rename_columns(self, names):
        names.extend(["T", "Sigma", "Z"])
        self.df.columns = names

    def drop_columns(self, columns, by_index=False):
        # TODO (dmt): Don't drop inplace.
        if not by_index:
            self.df.drop(columns=columns, inplace=True)
        else:
            assert all([isinstance(column, int) for column in columns])
            names = [self.df.columns[index] for index in columns]
            self.df.drop(columns=names, inplace=True)

    def labeled(self):
        return not self.df["Z"].nunique() == 1

    def as_numpy_array(self, effective=True):
        return self.df[self.columns(effective=effective)].values

    def set_column(self, name, values):
        if isinstance(values, list):
            assert len(values) == self.n_rows()
        self.df[name] = values

    def new_learnblock(self, values, columns, index, origin):
        data_frame = pd.DataFrame(data=values, columns=columns, index=index)
        block = PandasBlock(data_frame, origin=origin)
        return block

    def new_block_from(self, fundament, columns=None, time_column=False,
                       index=False):
        assert isinstance(fundament, list)

        if index:
            assert all([isinstance(index, int) for index in fundament])
            return PandasBlock(
                self.df.loc[fundament],
                self.relationship,
                self.subject,
                origin=self.origin)

        if time_column:
            assert all([isinstance(timestamp, int) for timestamp in fundament])
            return PandasBlock(
                self.df.loc[self.df["T"].isin(fundament)],
                self.relationship,
                self.subject,
                origin=self.origin)

    def get_column(self, name):
        return self.df[name].to_list()

    def get_row(self, index):
        return self.df.iloc[index].to_list()

    def sort_after(self, column, ascending=True):
        self.df.sort_values(by=[column], ascending=ascending, inplace=True)

    def get_duplicated_pairs(self, column_a, column_b):
        bool_series = self.df.duplicated(subset=[column_a, column_b])
        duplicates = self.df[bool_series]
        for i, j in zip(duplicates[column_a], duplicates[column_b]):
            yield i, j

    def construct_from_two_pairs(self, **kwargs):
        t, z, sigma = kwargs.get("T"), kwargs.get("Z"), kwargs.get("Sigma")

        if t is not None and z is not None:
            df = self.df.loc[(self.df["T"] == t) & (self.df["Z"] == z)]
        elif t is not None and sigma is not None:
            df = self.df.loc[(self.df["T"] == t) & (self.df["Sigma"] == sigma)]
        else:
            df = self.df.loc[(self.df["Z"] == z) & (self.df["Sigma"] == sigma)]

        return PandasBlock(df, self.relationship, origin=self.origin)

    def get_overlapping(self, block, drop_not=("T", "Z", "Z_other")):
        if len(block) < len(self):
            biggest = block
            block = self
        else:
            biggest = self
            block = block

        join = biggest.df.join(
            block.df.set_index("T"),
            on="T",
            lsuffix="_other",
            how="inner")

        if any(join.duplicated(["T"])):
            empty_df = pd.DataFrame()
            return PandasBlock(empty_df, origin=self.origin, subject=self.subject)

        for column in join.columns:
            if column not in drop_not:
                join.drop(column, axis=1, inplace=True)
        join = join[["Z", "Z_other", "T"]]
        # TODO (dmt): Make purpose and origin a tuple.

        return PandasBlock(join, origin=self.origin, subject=self.subject)

    def overlapping_rows(self, block, subset=None):
        big_df = pd.concat([self.df, block.df], sort=False)
        overlapping_data_frame = big_df[big_df.duplicated(
            subset=subset, keep=False)].drop_duplicates(keep="first")
        return PandasBlock(overlapping_data_frame,
                           subject=self.subject,
                           origin=self.origin)

    def same_features_fusion(self, block):
        df = pd.concat([self.df, block.df], sort=False, join="inner")
        origin = self.origin+block.origin
        return PandasBlock(df, subject=self.subject, origin=origin)

